// Copyright 2021 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

#include "ash/quick_pair/repository/fast_pair/device_image_store.h"

#include "ash/quick_pair/common/logging.h"
#include "ash/shell.h"
#include "chromeos/services/bluetooth_config/public/cpp/device_image_info.h"
#include "components/prefs/pref_registry_simple.h"
#include "components/prefs/pref_service.h"
#include "components/prefs/scoped_user_pref_update.h"
#include "ui/base/webui/web_ui_util.h"

namespace ash {
namespace quick_pair {

// Alias DeviceImageInfo for convenience.
using chromeos::bluetooth_config::DeviceImageInfo;

// static
constexpr char DeviceImageStore::kDeviceImageStorePref[];

// static
void DeviceImageStore::RegisterLocalStatePrefs(PrefRegistrySimple* registry) {
  registry->RegisterDictionaryPref(kDeviceImageStorePref);
}

DeviceImageStore::DeviceImageStore() = default;

DeviceImageStore::~DeviceImageStore() = default;

void DeviceImageStore::SaveDeviceImages(
    const std::string& model_id,
    DeviceMetadata* device_metadata,
    SaveDeviceImagesCallback on_images_saved_callback) {
  DCHECK(device_metadata);
  DeviceImageInfo& images = model_id_to_images_[model_id];

  if (images.default_image().empty() && !device_metadata->image().IsEmpty()) {
    SaveImageAsBase64(model_id, DeviceImageType::kDefault,
                      std::move(on_images_saved_callback),
                      device_metadata->image());
  } else {
    QP_LOG(WARNING) << __func__ << ": No default device image to save.";
    std::move(on_images_saved_callback).Run(SaveDeviceImagesResult::kFailure);
  }
}

bool DeviceImageStore::PersistDeviceImages(const std::string& model_id) {
  chromeos::bluetooth_config::DeviceImageInfo& images =
      model_id_to_images_[model_id];

  if (!DeviceImageInfoHasImages(images)) {
    QP_LOG(WARNING)
        << __func__
        << ": Attempted to persist non-existent images for model ID: " +
               model_id;
    return false;
  }
  PrefService* local_state = Shell::Get()->local_state();
  if (!local_state) {
    QP_LOG(WARNING) << __func__ << ": No shell local state available.";
    return false;
  }
  DictionaryPrefUpdate device_image_store(local_state, kDeviceImageStorePref);
  // TODO(dclasson): Once we add TrueWireless support, need to modify this to
  // merge new & persisted images objects.
  if (!device_image_store->SetKey(model_id, images.ToDictionaryValue())) {
    QP_LOG(WARNING) << __func__
                    << ": Failed to persist images to prefs for model ID: " +
                           model_id;
    return false;
  }
  return true;
}

bool DeviceImageStore::EvictDeviceImages(const std::string& model_id) {
  PrefService* local_state = Shell::Get()->local_state();
  if (!local_state) {
    QP_LOG(WARNING) << __func__ << ": No shell local state available.";
    return false;
  }
  DictionaryPrefUpdate device_image_store(local_state, kDeviceImageStorePref);
  if (!device_image_store->RemoveKey(model_id)) {
    QP_LOG(WARNING) << __func__
                    << ": Failed to evict images from prefs for model ID: " +
                           model_id;
    return false;
  }
  return true;
}

absl::optional<chromeos::bluetooth_config::DeviceImageInfo>
DeviceImageStore::GetImagesForDeviceModel(const std::string& model_id) {
  // Lazily load saved images from prefs the first time we get an image.
  if (!loaded_images_from_prefs_) {
    loaded_images_from_prefs_ = true;
    LoadPersistedImagesFromPrefs();
  }

  chromeos::bluetooth_config::DeviceImageInfo& images =
      model_id_to_images_[model_id];

  if (!DeviceImageInfoHasImages(images)) {
    return absl::nullopt;
  }
  return images;
}

void DeviceImageStore::LoadPersistedImagesFromPrefs() {
  PrefService* local_state = Shell::Get()->local_state();
  if (!local_state) {
    QP_LOG(WARNING) << __func__ << ": No shell local state available.";
    return;
  }
  const base::Value* device_image_store =
      local_state->GetDictionary(kDeviceImageStorePref);
  for (std::pair<const std::string&, const base::Value&> record :
       device_image_store->DictItems()) {
    absl::optional<chromeos::bluetooth_config::DeviceImageInfo> images =
        chromeos::bluetooth_config::DeviceImageInfo::FromDictionaryValue(
            record.second);
    if (!images) {
      QP_LOG(WARNING) << __func__
                      << ": Failed to load persisted images from prefs.";
      continue;
    }
    model_id_to_images_[record.first] = images.value();
  }
}

bool DeviceImageStore::DeviceImageInfoHasImages(
    const DeviceImageInfo& images) const {
  return !images.default_image_.empty() || !images.left_bud_image_.empty() ||
         !images.right_bud_image_.empty() || !images.case_image_.empty();
}

void DeviceImageStore::SaveImageAsBase64(
    const std::string& model_id,
    DeviceImageType image_type,
    SaveDeviceImagesCallback on_images_saved_callback,
    gfx::Image image) {
  // Encode the image as a base64 data URL.
  std::string encoded_image = webui::GetBitmapDataUrl(image.AsBitmap());

  // Save the image in the correct path.
  if (image_type == DeviceImageType::kDefault) {
    model_id_to_images_[model_id].default_image_ = encoded_image;
  } else {
    QP_LOG(WARNING) << __func__
                    << ": Can't save device image to invalid image field.";
    std::move(on_images_saved_callback).Run(SaveDeviceImagesResult::kFailure);
    return;
  }

  // Once successfully saved, return success.
  std::move(on_images_saved_callback).Run(SaveDeviceImagesResult::kSuccess);
}

}  // namespace quick_pair
}  // namespace ash
